/*
 * Copyright (c) 2013, 
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Nenad Corbic <ncorbic@sangoma.com>
 * 		 Pushkar Singh <psingh@sangoma.com>
 *
 */

#ifndef __SAMPLE_ISDN_TAPPING_H__
#define __SAMPLE_ISDN_TAPPING_H__

/********************************************************************************/
/*                                                                              */
/*                              Library Inclusion                               */
/*                                                                              */
/********************************************************************************/
#include <stdarg.h>
#include <ctype.h>
#ifdef __WINDOWS__
#include <io.h>
#include <sys/stat.h>
#endif
#include <fcntl.h>
#include "libsangoma_freetdm.h"

/********************************************************************************/
/*                                                                              */
/*                         Structure Declearation                               */
/*                                                                              */
/********************************************************************************/
typedef struct span_data {
	int init;
	int inuse;
	ftdm_span_t *ftdm_span;
	void *sangoma_port_obj;
	span_config_t span_conf; /* Application specified span configuration */
	sng_wp_config_t wp_conf; /* Application specified wanpipe configuration */
} span_data_t;

typedef enum {
	CALL_DIRECTION_INBOUND,
	CALL_DIRECTION_OUTBOUND,
	CALL_DIRECTION_NONE
} tap_call_direction_t;

typedef struct {
	uint32_t chan_id;		/* Channel Id on which message is received */
	uint32_t span_id;		/* Span Id on which call is received */
	tap_call_direction_t call_dir;	/* Gives Call direction i.e. incoming or outgoing */
	int call_ref;			/* Call reference val of the Call*/
	uint8_t call_con;		/* Gives information if call is connected or it is a missed call */
	char calling_num[FTDM_DIGITS_LIMIT]; /* Gives Calling Number */
	char called_num[FTDM_DIGITS_LIMIT];  /* Gives Called Number */
} isdn_tap_msg_info_t;

/********************************************************************************/
/*                                                                              */
/*                                     MACROS                                   */
/*                                                                              */
/********************************************************************************/
#define TAP_SUCCESS 0
#define TAP_SKIP    1
#define TAP_FAILURE -1
#define SNGISDN_TAP_EVENT_POLL_RATE 100

/* TAPPING CARDTYPE DECODE **********************************************************/
#define SNG_ISDN_TAP_CARD_TYPE(card_type) \
((card_type) == A101_ADPTR_1TE1)	?"A101 single E1/T1 channels" : \
((card_type) == A101_ADPTR_2TE1)      	?"A101 bi E1/T1 channels" : \
((card_type) == A104_ADPTR_4TE1)        ?"A104 quad line E1/T1" : \
((card_type) == A108_ADPTR_8TE1)        ?"A108 octa E1/T1 channels" : \
((card_type) == A116_ADPTR_16TE1)	?"A116 16 channels E1/T1" : \
((card_type) == AFT_ADPTR_ISDN)		?"A 500 BRI card" : \
((card_type) == AFT_ADPTR_W400)         ?"W400 GSM" : \
"Unknown"

/* ISDN CALL DIRECTION *************************************************************/
#define ISDN_CALL_DIRECTION(call_dir) \
	((call_dir) == CALL_DIRECTION_INBOUND)		?"Incoming Call" : \
	((call_dir) == CALL_DIRECTION_OUTBOUND)		?"Outgoing Call" : \
"Invalid Call Direction"

#endif							/* __SAMPLE_ISDN_TAPPING_H__ */
