/*
 * Copyright (c) 2014, 
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Nenad Corbic <ncorbic@sangoma.com>
 * 		 Pushkar Singh <psingh@sangoma.com>
 *
 */

/********************************************************************************/
/*                                                                              */
/*                           libraries inclusion                                */
/*                                                                              */
/********************************************************************************/
#include "sample_isup_tapping.h"

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
#ifndef __WINDOWS__
void on_sigint(int signal);
void on_sigusr(int signal);
#endif

int exit_sample_app(int max_wait);
int map_tapping_spans(void *sangoma_port_obj, int wanpipe_num, span_config_t * span_conf);
int open_audio_file(uint32_t chan_id, uint32_t span_id, void *obj);

/* Set SDK_DEBUGGING since this is a sample application,
 * This should be turned off in Production */
#define SDK_DEBUGGING 1

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;

span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE + 1] = { {0} };
sangoma_freetdm_init_t sample_app_ftdm_init;

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

/**********************************************************************************
 * Fun   : on_sigusr()
 *
 * Desc  : In case if there is any iterruption by user in order then stop the 
 * 	   application and exit
 *
 * Ret   : void
**********************************************************************************/
#ifndef __WINDOWS__
void on_sigusr(int signal)
{
	switch (signal) {
	case SIGUSR1:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		break;

	case SIGUSR2:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		break;
	}

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA Tapping Module. Application is getting closed\n");
	getchar();
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");

	exit(EXIT_FAILURE);
}

/**********************************************************************************
 * Fun   : on_sigint()
 *
 * Desc  : In case if SIGINT is received then stop the application and exit
 *
 * Ret   : void
**********************************************************************************/
void on_sigint(int signal)
{
	g_running = 0;

	ftdm_log(FTDM_LOG_INFO, "SIGINT recieved ...\n");

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA Tapping Module. Application is getting closed\n");
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");
	getchar();

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	fprintf(stdout, "\n\n\n=========================================================\n");
	fprintf(stdout, "                           GOOD BYE                            \n");
	fprintf(stdout, "=========================================================\n");

	exit(EXIT_FAILURE);
}
#endif

/****************************************
 FreeTDM Event Callbacks
****************************************/
/**********************************************************************************
 * Fun   : sample_app_hwprobe_callback()
 *
 * Desc  : Stop all the ports one by one then save the configuration and then start
 * 	   the respective wanpipe/port. Wait for some time in order to get the
 * 	   wanpipe in connected state and create respecteive FreeTDM spans
 *
 * Ret   : FTDM_SUCCESS=0 | FTDM_FAIL | -1
**********************************************************************************/
int sample_app_hwprobe_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int err;

	/* Filter out all other cards just look for cards Tapping is supported */
	switch (card_type) {
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
		break;

	default:
		ftdm_log(FTDM_LOG_INFO, "%s card (%d) does not support ISUP Tapping\n", SNG_ISUP_TAP_CARD_TYPE(card_type),
				 card_type);
		return FTDM_FAIL;
	}

	/* NOTE: Below condition is only for testing purpose
	 * if condition only for testing specific ports ..real app should
	 * not have such condition */
	if (port_id != 13 && port_id != 14 && port_id != 15 && port_id != 16) {
		return FTDM_FAIL;
	}

	SPANS[port_id].init = 1;
	SPANS[port_id].sangoma_port_obj = sangoma_port_obj;

	/* NOTE: Below condition is only for testing purpose
	 * if condition is only to configure respective pure voice span as per testing requirement..
	 * real app shoul have /not conditions as per customers requirement */
	/* Re-Configuring ISUP signalling span to pure voice span as per requirement */
	if (port_id == 15 || port_id == 16) {
		err = sangoma_set_default_isup_pure_voice_span_config(sangoma_port_obj);
		if (!err) {
			ftdm_log(FTDM_LOG_DEBUG, "Configured wanpipe wp%d for pure voice span\n", port_id);
		}
	}

	/* Stop the respective port/wanpipe */
	err = sangoma_config_port_stop(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
	}

	/* Save the respective port/wanpipe configuration */
	err = sangoma_config_port_save(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
		return -1;
	}

	/* Start the respective port/wanpipe */
	err = sangoma_config_port_start(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to start sangoma port config\n");
		return -1;
	}

	/* NOTE:  Below condition only for testing purpose
	 * to specify the pure voice spans..
	 * real app will call the api based on there configuration of span i.e. pure voice Vs signalling */
	/* Create spans in FreeTDM */
	if (port_id == 15 || port_id == 16) {
		err = sangoma_freetdm_create_span(sangoma_port_obj, &SPANS[port_id].ftdm_span, SNG_SS7_ISUP_PURE_VOICE_TAPPING_MODE);
	} else {
		err = sangoma_freetdm_create_span(sangoma_port_obj, &SPANS[port_id].ftdm_span, SNG_SS7_ISUP_SIG_TAPPING_MODE);
	}

	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to create freetdm span %i\n", port_id);
		return -1;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : sample_app_span_start_callback()
 *
 * Desc  : For Tapping one needs to carry config and start span once all spans
 * 	   gets created as for tapping purpose one have to provide the peerspan
 * 	   for each span configured. This function mainly configures spans and 
 * 	   start the same.
 *
 * Ret   : FTDM_SUCCESS | -1 | ret
**********************************************************************************/
int sample_app_span_start_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int ret = 0;
	span_config_t span_conf;

	/* Filter out all other cards just look for cards Tapping is supported */
	ret = map_tapping_spans(sangoma_port_obj, port_id, &span_conf);
	if (ret != FTDM_SUCCESS) {
		return ret;
	}
	
	/* Loads the configuration with respect to TAPPING module */
	if (sangoma_freetdm_span_config_and_start(sangoma_port_obj, SPANS[port_id].ftdm_span, &span_conf, SNG_SS7_ISUP_SIG_TAPPING_MODE)) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to load and configure freetdm span %i\n", port_id);
		return -1;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : tapping_main_menu()
 *
 * Desc  : Provides option in order to stop the application at any point of time
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
int tapping_main_menu(void)
{
	char choice[6];

	ftdm_log(FTDM_LOG_INFO, "=========================================================\n");
	ftdm_log(FTDM_LOG_INFO, "	    SANGOMA ISUP TAPPING MODULE STARTED           \n");
	ftdm_log(FTDM_LOG_INFO, "=========================================================\n");
	ftdm_log(FTDM_LOG_INFO, "Please enter 1 at any point of time if you want to exit from application \n");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			return FTDM_FAIL;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Exit from application
		 */
		switch (choice[0]) {
		case '1':
			ftdm_log(FTDM_LOG_INFO, "SANGOMA Tapping Sample Application is getting closed...\n");

			/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
			 * the configired wanpipes/hwports */
			exit_sample_app(20);

			fprintf(stdout, "\n\n\n=========================================================\n");
			fprintf(stdout, "                      GOOD BYE                            \n");
			fprintf(stdout, "=========================================================\n");

			return FTDM_SUCCESS;

		default:
			break;
		}
	}

	return FTDM_FAIL;
}

/**********************************************************************************
 * Fun   : start_tapping()
 *
 * Desc  : Starts tapping the call on the respective span and channel on which 
 * 	   call is established
 *
 * Ret   : NULL
**********************************************************************************/
static void *start_tapping(ftdm_thread_t * me, void *obj)
{
	uint8_t frame[1024];
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	ftdm_channel_t *chan = NULL;
	ftdm_span_t *span = NULL;
	ftdm_size_t len;
	uint32_t span_id = sigmsg->span_id;
	uint32_t chan_id = sigmsg->chan_id;
	ftdm_caller_data_t *caller_data = NULL;
	char timestamp[80];
	char path[512];
	ftdm_size_t dlen=0x00;
	int fd = -1;

	memset(path, 0 , sizeof(path));
	memset(timestamp, 0 , sizeof(timestamp));

	ftdm_log(FTDM_LOG_DEBUG, "Enter: start_tapping on span_id[%d] chanid[%d] \n", span_id, chan_id);

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid signalling message\n");
		goto done;
	}

	if (!span_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Span ID\n");
		goto done;
	}

	ftdm_span_find(span_id, &span);

	/* Send request to Freetdm to start tracing the media */
	if (!chan_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel ID\n");
		goto done;
	}

	if (!span) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid span found\n");
		goto done;
	}
	chan = ftdm_span_get_channel(span, chan_id);

	if (!chan) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid channel found\n");
		goto done;
	}

	ftdm_log(FTDM_LOG_DEBUG, "Start tapping on span %d and channel %d\n", span_id, chan_id);

	sangoma_get_timestamp_str(timestamp);

	caller_data = ftdm_channel_get_caller_data(chan);

	if (caller_data) {
		if (strlen(caller_data->dnis.digits)) {
			sprintf(path, "%s-%d-%d-%s", timestamp, span_id, chan_id, caller_data->dnis.digits);
		} else {
			sprintf(path, "%s-%d-%d", timestamp, span_id, chan_id);
		}
	} else {
		sprintf(path, "%s-%d-%d", timestamp, span_id, chan_id);
	}

	if (!strlen(path)) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid Path\n");
		goto done;
	}

	if ((fd = open_audio_file(chan_id, span_id, path)) == -1) {
		ftdm_log(FTDM_LOG_DEBUG, "Failed to start audio tracing on span[%d]-chan[%d]\n", span_id, chan_id);
		goto done;
	}

	/* Start reading the media, and write to a file in order to tap the call */ 
	while (ftdm_channel_call_check_answered(chan)) {
		ftdm_wait_flag_t flags = FTDM_READ;

		if (ftdm_channel_wait(chan, &flags, -1) == FTDM_FAIL) {
			break;
		}
		len = sizeof(frame);

		if (flags & FTDM_READ) {
			if (ftdm_channel_read(chan, frame, &len) == FTDM_SUCCESS) {
				/* Passive Tapping application should not write to channel, only read and write the data
				 * to a file */
				dlen = len;
				if ((ftdm_size_t)write(fd, frame, (int)dlen) != dlen) {
					ftdm_log(FTDM_LOG_WARNING, "Raw input trace failed to write all of the %"FTDM_SIZE_FMT" bytes\n", dlen);
				}
			}
		}
	}

	if (fd >= 0) {
		close(fd);
		fd = -1;
	}

  done:
	ftdm_free(sigmsg);
	return NULL;
}

/**********************************************************************************
 * Fun   : app_freetdm_event_queue_handler()
 *
 * Desc  : Starts respective io operation w.r.t different event received 
 *
 * Ret   : NULL
**********************************************************************************/
static void *app_freetdm_event_queue_handler(ftdm_thread_t * me, void *obj)
{
	ftdm_interrupt_t 	*ftdm_sangoma_isup_int[1];
	ftdm_status_t 		ret_status;
	int32_t 		sleep = SNGISUP_TAP_EVENT_POLL_RATE;
	ftdm_sigmsg_t 		*sigmsg = NULL;

	if (ftdm_queue_get_interrupt(sample_app_ftdm_init.event_msg_queue, ftdm_sangoma_isup_int) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_CRIT, "Failed to get a event interrupt for app_freetdm_event_queue_handler thread !!\n");
		goto exit;
	}

	while (g_running) {
		ret_status = ftdm_interrupt_multiple_wait(ftdm_sangoma_isup_int, 1, sleep);

		/* Find out why we returned from the interrupt queue */
		switch (ret_status) {
			case FTDM_SUCCESS:  /* There was a state change on the span */
				{
					/* It can possible that FreeTDM can pump lots of messages to queue, thus read all messages 
					 * untill queue is empty */
					while (sigmsg = ftdm_queue_dequeue(sample_app_ftdm_init.event_msg_queue)) {
						ftdm_sigmsg_t *sigmsg_cpy = NULL;

						switch (sigmsg->event_id) {
							case FTDM_SIGEVENT_START:
								{
									sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
									memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
									ftdm_log(FTDM_LOG_DEBUG, "SIG_START event received. Starts io data operation for [s%dc%d]. \n",
											sigmsg->span_id, sigmsg->chan_id);

									/* Incoming Call. Proceed to answer the call and then
									   launch a media thread that will perform I/O on the channel */
									ftdm_channel_call_answer(sigmsg->channel);

									if (ftdm_thread_create_detached(start_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
										ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
												sigmsg->span_id, sigmsg->chan_id);
										break;
									}

									break;
								}

							case FTDM_SIGEVENT_UP:
								{
									sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
									memcpy(sigmsg_cpy, sigmsg, sizeof(ftdm_sigmsg_t));
									ftdm_log(FTDM_LOG_DEBUG, "SIG_UP event received. Starts io data operation for [s%dc%d]. \n",
											sigmsg->span_id, sigmsg->chan_id);

									/* Outbound call has been answered, proceed to launch a meda thread that will
									   perform I/O on the channel */
									if (ftdm_thread_create_detached(start_tapping, sigmsg_cpy) != FTDM_SUCCESS) {
										ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
												sigmsg->span_id, sigmsg->chan_id);
										break;
									}
									break;
								}

							default:
								break;
						}

						/* sample app owns the event buffer. It must be deallocated here
						   otherwise there is possibility of memory leak */
						ftdm_safe_free(sigmsg);
					} /* while (dequeue) */
				}
				break;

			case FTDM_TIMEOUT:
				/* twiddle */
				break;

			case FTDM_FAIL:
				ftdm_log(FTDM_LOG_ERROR, "ftdm_interrupt_wait returned error!\n");
				break;

			default:
				ftdm_log(FTDM_LOG_ERROR, "ftdm_interrupt_wait returned with unknown code\n");
				break;
		} /* switch (ret_status) */
	} /* while */

exit:
	ftdm_log(FTDM_LOG_CRIT, "app_freetdm_event_queue_handler monitor thread stopping due to error.\n");
	return NULL;
}

/**********************************************************************************
 * Fun   : main()
 *
 * Desc  : Configures all ports and spans as required by user and provide required
 * 	   platform for tapping calls 
 *
 * Ret   : err | FTDM_SUCCESS
**********************************************************************************/
int main(void)
{
	/* local variables initialization */
	int max_wait = 20;
	int err;
	int detected = 0;
	int configured = 0;
	sangoma_hw_config_reg_t reg;

	memset(&reg, 0, sizeof(reg));

#ifndef __WINDOWS__
	/* Signal handler in order to handle different signal */
	signal(SIGINT, on_sigint);
	signal(SIGTERM, on_sigint);
	signal(SIGUSR1, on_sigusr);
	signal(SIGUSR2, on_sigusr);
#endif

	/* Initialize global FreeTDM init structure */
	memset(&sample_app_ftdm_init, 0, sizeof(sample_app_ftdm_init));

	/* Set basic debuggin parameters */
#ifdef SDK_DEBUGGING
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_DEBUG;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_ON_ASSERT;
#else
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_INFO;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_NEVER;
#endif

	/* Initilize FreeTDM Framework 
	   Global Varialbes and FreeTDM Logger */
	err = sangoma_freetdm_init(&sample_app_ftdm_init);
	if (err) {
		return err;
	}

	/* Initialize Sangoma Config SDK */
	reg.user_obj = NULL;
	reg.hwprobe_callback = sample_app_hwprobe_callback;
	reg.hwprobe_callback_post = sample_app_span_start_callback;
	err = sangoma_config_reg(&reg);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Sangoma Config SDK Initialization Error\n");
		exit_sample_app(max_wait);
	}

	/* Create a global FreeTDM event handling queue.
	 * FreeTDM event callback function is not allowed to exectue FreeTDM API commands.
	 * FreeTDM event callback fucntion will queue events, and the app event handler thread
	 * will consume and handle FreeTDM events.  
	 * The app event handler is allowed to run FreeTDM api commands directly */
	if (ftdm_thread_create_detached(app_freetdm_event_queue_handler, NULL) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create message queue handler thread. \n");
		exit_sample_app(max_wait);
	}

	/* Detect Sangoma Ports 
	   For each detected port run the sample_app_hwprobe_callback
	 */
	err = sangoma_config_hwprobe(&detected, &configured, SNG_SS7_ISUP_SIG_TAPPING_MODE);

	ftdm_log(FTDM_LOG_INFO, "Sangoma SDK Detected %d Configured %d\n", detected, configured);

	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error Failed to detect sangoma hardware \n");

		/* We cannot exit here, as the thread is launched */
		exit_sample_app(max_wait);
	}
	/* At this point all sangoma hw ports have been configured and started.
	   And FreeTDM Spans have been created and started on each hw port */
	ftdm_sleep(1000);

	/* User drive main menu. 
	 * Test features for the Sangoma TDM SDK 
	 */
	ftdm_log(FTDM_LOG_INFO, "Sangoma ISUP TAPPING MODULE IN ACTION\n\n\n");
	while (tapping_main_menu()) {
		continue;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : exit_sample_app()
 *
 * Desc  : Stops all FTDM as well as application thread and deallocate the memory
 * 	   stop all the configured ports and exit from application.
 *
 * Ret   : NULL
**********************************************************************************/
int exit_sample_app(int max_wait)
{

	/* Stop all threads. 
	 * TODO: use join instead */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		ftdm_log(FTDM_LOG_INFO, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	/* Stop FreeTDM Freamwork and all Spans */
	ftdm_global_destroy();

	/* Stop all Sangoma hw ports */
	sangoma_stop_all_wanpipe();

	ftdm_sleep(1000);

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : map_tapping_spans()
 *
 * Desc  : Detect the port number and configure span w.r.t tapping module, also
 * 	   peer span is selected depending up on the port and wanpipe number w.r.t
 * 	   that particular port 
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL | -1
**********************************************************************************/
int map_tapping_spans(void *sangoma_port_obj, int wanpipe_num, span_config_t * span_conf)
{
	char peer_span_name[20];
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int port_num = sangoma_config_port_hwinfo_get_port_number(sangoma_port_obj);

	memset(span_conf, 0, sizeof(span_config_t));

	/* A102 has 2 ports - thus case 1 and 2 will match
	   A104 has 4 ports - thus case 1,2,3,4 will match
	   A108 has 8 ports */

	if ((port_num % 2) != 0) {
		wanpipe_num++;
	} else {
		wanpipe_num--;
	}

	sprintf(peer_span_name, "wp%d", wanpipe_num);

	switch (card_type) {
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
		{
			switch (port_num) {
			case 1:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "1");
				strcpy(span_conf->config.isup.is_voice, "0");
				strcpy(span_conf->config.isup.voice_span, "15");
				break;

			case 2:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "1");
				strcpy(span_conf->config.isup.is_voice, "0");
				strcpy(span_conf->config.isup.voice_span, "16");
				break;

			case 3:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "33");
				strcpy(span_conf->config.isup.is_voice, "1");
				strcpy(span_conf->config.isup.voice_span, "0");
				break;

			case 4:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "33");
				strcpy(span_conf->config.isup.is_voice, "1");
				strcpy(span_conf->config.isup.voice_span, "0");
				break;

			case 5:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "1");
				strcpy(span_conf->config.isup.is_voice, "1");
				break;

			case 6:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "1");
				strcpy(span_conf->config.isup.is_voice, "1");
				break;

			case 7:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "1");
				strcpy(span_conf->config.isup.is_voice, "1");
				break;

			case 8:
				strcpy(span_conf->config.isup.peer_span, peer_span_name);
				strcpy(span_conf->config.isup.switch_type, "itu00");
				strcpy(span_conf->config.isup.cicbase, "1");
				strcpy(span_conf->config.isup.is_voice, "1");
				break;

			default:
				ftdm_log(FTDM_LOG_ERROR, "Invalid port id = %d\n", port_num);
				return -1;
				break;
			}
		}
		break;

	default:
		ftdm_log(FTDM_LOG_ERROR, "%s card (%d) does not support ISUP Tapping\n", SNG_ISUP_TAP_CARD_TYPE(card_type),
				 card_type);
		return FTDM_FAIL;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : open_audio_file()
 *
 * Desc  : Open the the file to which raw data needs to  be written in order to tap
 * 	   the call.
 *
 * Ret   : -1 | fd
**********************************************************************************/
int open_audio_file(uint32_t chan_id, uint32_t span_id, void *obj)
{
	int fd = 0;
	char *path = (char *)obj;

	if (!path) {
		ftdm_log(FTDM_LOG_ERROR, "Cannot Open path as it is empty\n");
		return -1;
	}

	if ((fd = open(path, O_WRONLY | O_CREAT | O_TRUNC
					| FTDM_O_BINARY, S_IRUSR | S_IWUSR)) < 0 ) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to open file [%s]\n", path);
		return -1;
	}
	ftdm_log(FTDM_LOG_DEBUG, "Start Tracing Call on channel %d:%d to [%s]\n", span_id, chan_id, path);
	return fd;
}
