/*
 * Copyright (c) 2014, 
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Nenad Corbic <ncorbic@sangoma.com>
 * 		 Pushkar Singh <psingh@sangoma.com>
 *
 */

/********************************************************************************/
/*                                                                              */
/*                           libraries inclusion                                */
/*                                                                              */
/********************************************************************************/
#include "sample_mtp2_api.h"

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
#ifndef __WINDOWS__
void on_sigint(int signal);
void on_sigusr(int signal);
#endif

int exit_sample_app(int max_wait);
int config_spans(void *sangoma_port_obj, int wanpipe_num, span_config_t * span_conf);

/* Set SDK_DEBUGGING since this is a sample application,
 * This should be turned off in Production */
#define SDK_DEBUGGING 1

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/
int g_running = 1;
int g_num_threads = 0;

span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE + 1] = { {0} };
sangoma_freetdm_init_t sample_app_ftdm_init;
static ftdm_hash_t *ss7_configs;
sng_isup_t sng_isup_conf;

/********************************************************************************/
/*                                                                              */
/*                              implementation                                  */
/*                                                                              */
/********************************************************************************/

/**********************************************************************************
 * Fun   : on_sigusr()
 *
 * Desc  : In case if there is any iterruption by user in order then stop the 
 * 	   application and exit
 *
 * Ret   : void
**********************************************************************************/
#ifndef __WINDOWS__
void on_sigusr(int signal)
{
	switch (signal) {
	case SIGUSR1:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		print_frames();
		break;

	case SIGUSR2:
		ftdm_log(FTDM_LOG_INFO, "User interrupted... \n");
		print_status();
		break;
	}

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA MTP2 API Module. Application is getting closed\n");
	getchar();
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	ftdm_log(FTDM_LOG_INFO, "\n\n\n=========================================================\n");
	ftdm_log(FTDM_LOG_INFO, "                           GOOD BYE                            \n");
	ftdm_log(FTDM_LOG_INFO, "=========================================================\n");

	exit(EXIT_FAILURE);
}

/**********************************************************************************
 * Fun   : on_sigint()
 *
 * Desc  : In case if SIGINT is received then stop the application and exit
 *
 * Ret   : void
**********************************************************************************/
void on_sigint(int signal)
{
	g_running = 0;

	ftdm_log(FTDM_LOG_INFO, "SIGINT recieved ...\n");

	ftdm_log(FTDM_LOG_INFO, "Exiting SANGOMA MTP2 API Module. Application is getting closed\n");
	ftdm_log(FTDM_LOG_INFO, "Please press enter... \n");
	getchar();

	/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
	 * the configired wanpipes/hwports */
	exit_sample_app(20);

	ftdm_log(FTDM_LOG_INFO, "\n\n\n=========================================================\n");
	ftdm_log(FTDM_LOG_INFO, "                           GOOD BYE                            \n");
	ftdm_log(FTDM_LOG_INFO, "=========================================================\n");

	exit(EXIT_FAILURE);
}
#endif

/****************************************
 FreeTDM Event Callbacks
****************************************/
/**********************************************************************************
 * Fun   : sample_app_hwprobe_callback()
 *
 * Desc  : Stop all the ports one by one then save the configuration and then start
 * 	   the respective wanpipe/port. Wait for some time in order to get the
 * 	   wanpipe in connected state and create respecteive FreeTDM spans
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL | -1
**********************************************************************************/
int sample_app_hwprobe_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int err;

	/* Filter out all other cards just look for cards which supports mtp2 api */
	switch (card_type) {
	case A101_ADPTR_2TE1:
	case A104_ADPTR_4TE1:
	case A108_ADPTR_8TE1:
	case A116_ADPTR_16TE1:
		break;

	default:
		ftdm_log(FTDM_LOG_INFO, "%s card (%d) does not support MTP2 API\n", SNG_MTP2_API_CARD_TYPE(card_type),
				 card_type);
		return FTDM_FAIL;
	}

	if ((port_id != 13) && (port_id != 14)) {
		return FTDM_FAIL;
	}

	SPANS[port_id].init = 1;
	SPANS[port_id].sangoma_port_obj = sangoma_port_obj;

	/* Stop the respective port/wanpipe */
	err = sangoma_config_port_stop(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
	}

	/* Save the respective port/wanpipe configuration */
	err = sangoma_config_port_save(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to save sangoma port config\n");
		return -1;
	}

	/* Start the respective port/wanpipe */
	err = sangoma_config_port_start(sangoma_port_obj);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to start sangoma port config\n");
		return -1;
	}

	/* Create spans in FreeTDM */
	err = sangoma_freetdm_create_span(sangoma_port_obj, &SPANS[port_id].ftdmspan, SNG_SS7_MTP2_API_MODE_LSL);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to create freetdm span %i\n", port_id);
		return -1;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : sample_app_span_start_callback()
 *
 * Desc  : Start the spans once the wanpipe configuration is done.
 *
 * Ret   : FTDM_SUCCESS | -1 | ret
**********************************************************************************/
int sample_app_span_start_callback(void *sangoma_port_obj, int port_id, void *user_obj)
{
	int ret = 0;
	span_config_t span_conf;

	if (port_id == 13) {
		ftdm_log(FTDM_LOG_INFO, "Waiting.... In order to get wanpipe in connected state\n");
		/* Due to test app scenario which is configuring 2 spans and sending/receiving data and validating that data, we need
		 * to make sure wanpipes are up before we start sending data...
		 * If we dont wait then it might possible that 1 end can come up and start sending data while other end(span)
		 * still down which will leads to mismatch of data and app will stop due to validation failure  */
		sleep(31);
	}
	/* Filter out all other cards just look for cards which supports mtp2 api */
	ret = config_spans(sangoma_port_obj, port_id, &span_conf);
	if (ret != FTDM_SUCCESS) {
		return ret;
	}
	
	/* Check if no node is found that needs to be configured */
	if (span_conf.config.mtp2_api.ss7confnode == NULL) {
		ftdm_log(FTDM_LOG_ERROR, "No node found that needs to be configured. Please check your configuration\n");
		return -1;
	}

	/* Loads the configuration with respect to MTP2 API module */
	if (sangoma_freetdm_span_config_and_start(sangoma_port_obj, SPANS[port_id].ftdmspan, &span_conf, SNG_SS7_MTP2_API_MODE_LSL)) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Failed to load and configure freetdm span %i\n", port_id);
		return -1;
	}

	SPANS[port_id].inuse = 1;
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : mtp2_api_main_menu()
 *
 * Desc  : Provides option in order to stop the application at any point of time
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL
**********************************************************************************/
int mtp2_api_main_menu(void)
{
	char choice[6];

	ftdm_log(FTDM_LOG_INFO, "=========================================================\n");
	ftdm_log(FTDM_LOG_INFO, "              SANGOMA MTP2 API MODULE STARTED            \n");
	ftdm_log(FTDM_LOG_INFO, "=========================================================\n");
	ftdm_log(FTDM_LOG_INFO, "1. Print Status\n");
	ftdm_log(FTDM_LOG_INFO, "2. Print Frames Rx/Tx Status\n");
	ftdm_log(FTDM_LOG_INFO, "3. Exit\n");
	ftdm_log(FTDM_LOG_INFO, "Enter your choice: ");

	if (NULL != fgets(choice, 6, stdin)) {
		if (2 < strlen(choice)) {
			return FTDM_FAIL;
		}

		choice[1] = '\0';

		/* Provides option to user to perform various operations like:
		 * 1) Exit from application
		 */
		switch (choice[0]) {
			case '1':
				print_status();
				break;

			case '2':
				print_frames();
				break;

			case '3':
				ftdm_log(FTDM_LOG_INFO, "SANGOMA MTP2 API Sample Application is getting closed...\n");

				print_frames();
				/* Stop all FTDM as well as application thread and deallocated the memory. Also, stop all
				 * the configired wanpipes/hwports */
				exit_sample_app(20);

				ftdm_log(FTDM_LOG_INFO, "\n\n\n=========================================================\n");
				ftdm_log(FTDM_LOG_INFO, "                      GOOD BYE                            \n");
				ftdm_log(FTDM_LOG_INFO, "=========================================================\n");

				return FTDM_SUCCESS;

			default:
				break;
		}
	}

	return FTDM_FAIL;
}

/**********************************************************************************
 * Fun   : Verify()
 *
 * Desc  : Verify the recieved data whether it is good
 *
 * Ret   : Null
**********************************************************************************/
static void *verify(ftdm_thread_t * me, void *obj)
{
	ftdm_sigmsg_t *sigmsg = (ftdm_sigmsg_t *) obj;
	ftdm_channel_t          *ftdmchan;
	chan_data_t             *chan;

	if (!sigmsg) {
		ftdm_log(FTDM_LOG_INFO, "Invalid signalling messages\n");
		goto done;
	}

	ftdmchan = sigmsg->channel;
	chan = (chan_data_t*)ftdm_channel_get_private(ftdmchan);

	verify_receive_data(chan, sigmsg->raw.data, sigmsg->raw.len);

done:
	ftdm_safe_free(sigmsg->raw.data);
	ftdm_safe_free(sigmsg);
	return NULL;
}

/**********************************************************************************
 * Fun   : app_freetdm_event_queue_handler()
 *
 * Desc  : Starts respective io operation w.r.t different event received
 *
 * Ret   : NULL
**********************************************************************************/
static void *app_freetdm_event_queue_handler(ftdm_thread_t * me, void *obj)
{
	ftdm_sigmsg_t 		*sigmsg;
	unsigned int 		span_id, chan_id;
	ftdm_channel_t 		*ftdmchan;
	chan_data_t 		*chan;
	ftdm_interrupt_t        *ftdm_sangoma_mtp2_int[1];
	ftdm_status_t           ret_status;

	if (ftdm_queue_get_interrupt(sample_app_ftdm_init.event_msg_queue, ftdm_sangoma_mtp2_int) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_CRIT, "Failed to get a event interrupt for app_freetdm_event_queue_handler thread !!\n");
		goto exit;
	}

	while (g_running) {
		ret_status = ftdm_interrupt_multiple_wait(ftdm_sangoma_mtp2_int, 1, 100);

		/* Find out why we returned from the interrupt queue */
		switch (ret_status) {
			case FTDM_SUCCESS:  /* There was a state change on the span */
				{
					while (sigmsg = ftdm_queue_dequeue(sample_app_ftdm_init.event_msg_queue)) {
						ftdm_sigmsg_t *sigmsg_cpy = NULL;

						ftdmchan = sigmsg->channel;

						span_id = ftdm_channel_get_ph_span_id(ftdmchan);
						chan_id = ftdm_channel_get_ph_id(ftdmchan);

						chan = (chan_data_t*)ftdm_channel_get_private(ftdmchan);

						if (!chan) {
							ftdm_log(FTDM_LOG_INFO, "Please wait as still chan is not being allocated\n");
							ftdm_safe_free(sigmsg);
							break;
						}

						switch (sigmsg->event_id) {
							case FTDM_SIGEVENT_IO_INDATA:
								/* We received an incoming frame */
								sigmsg_cpy = ftdm_malloc(sizeof(ftdm_sigmsg_t));
								memcpy(sigmsg_cpy, sigmsg, sizeof(*sigmsg));
								if (sigmsg->raw.data)
								{
									sigmsg_cpy->raw.data = ftdm_malloc(sigmsg->raw.len);
									memcpy(sigmsg_cpy->raw.data, sigmsg->raw.data, sigmsg->raw.len);
								}

								ftdm_log(FTDM_LOG_DEBUG, "Starts io data operation for [s%dc%d]. \n",
										sigmsg->span_id, sigmsg->chan_id);

								if (ftdm_thread_create_detached(verify, sigmsg_cpy) != FTDM_SUCCESS) {
									ftdm_log(FTDM_LOG_ERROR, "Not able to launch IO thread for [s%dc%d]. \n",
											sigmsg->span_id, sigmsg->chan_id);
									break;
								}
								break;

							case FTDM_SIGEVENT_RAW:
								{
									switch (sigmsg->ev_data.raw_id) {
										case SNGSS7_IND_RETRV_BSN:			/* Response to a SNGSS7_RAW_ID_REQ_RETRV_BSN  */
											break;
										case SNGSS7_IND_DATA_MORE:			/* Response to a SNGSS7_RAW_ID_REQ_RETRV_MSG, more frames to come */
											break;
										case SNGSS7_IND_DATA_NO_MORE:			/* Response to a SNGSS7_RAW_ID_REQ_RETRV_MSG, this is the last frame */
											break;
									}
								}
								break;

							case FTDM_SIGEVENT_SIGSTATUS_CHANGED:
								{
									ftdm_log(FTDM_LOG_INFO, "s%dc%d: Sig status changed to:%s\n", span_id, chan_id, ftdm_signaling_status2str(sigmsg->ev_data.sigstatus.status));

									switch (sigmsg->ev_data.sigstatus.status) {
										case FTDM_SIG_STATE_UP:
											chan->flags |= LINK_STATE_CONNECTED;
											chan->flags &= ~LINK_STATE_CONGESTED;

											switch (sigmsg->ev_data.sigstatus.raw_reason) {
												case SNGSS7_REASON_END_FLC:                     /* End of flow control */
													ftdm_log(FTDM_LOG_INFO, "s%dc%d: Flow control end\n", span_id, chan_id);
													break;
												case SNGSS7_REASON_REM_PROC_UP:                 /* Remote MTP2 processor is UP */
													ftdm_log(FTDM_LOG_INFO, "s%dc%d: Remote processor up\n", span_id, chan_id);
													break;
												default:
													ftdm_log(FTDM_LOG_INFO, "s%dc%d: MTP2 Link aligned\n", span_id, chan_id);
													break;
											}
											break;

										case FTDM_SIG_STATE_DOWN:
											{
												uint8_t raw_reason = sigmsg->ev_data.sigstatus.raw_reason;
												chan->flags &= ~LINK_STATE_CONNECTED;

												ftdm_log(FTDM_LOG_INFO, "s%dc%d: MTP2 Link down: \"%s\" \n", span_id, chan_id,
														(raw_reason == SNGSS7_REASON_SM) ?"Stack Manager requested DISC":
														(raw_reason == SNGSS7_REASON_SUERM) ?"SUERM threshold reached":
														(raw_reason == SNGSS7_REASON_ACK) ?"Excessive delay of acknowledgements from remote MTP2":
														(raw_reason == SNGSS7_REASON_TE) ?"Failure of Terminal equipment":
														(raw_reason == SNGSS7_REASON_BSN) ?"2 of 3 unreasonable BSN":
														(raw_reason == SNGSS7_REASON_FIB) ?"2 of 3 unreasonable FIB":
														(raw_reason == SNGSS7_REASON_CONG) ?"Excessive periods of congestion":
														(raw_reason == SNGSS7_REASON_LSSU_SIOS) ?"SIO/SIOS received in link state machine":
														(raw_reason == SNGSS7_REASON_TMR2_EXP) ?"Timer T2 expired":
														(raw_reason == SNGSS7_REASON_TMR3_EXP) ?"Timer T3 expired":
														(raw_reason == SNGSS7_REASON_LSSU_SIOS_IAC) ?"SIOS received during alignment":
														(raw_reason == SNGSS7_REASON_PROV_FAIL) ?"ERM threshold reached":
														(raw_reason == SNGSS7_REASON_TMR1_EXP) ?"Timer T1 expired":
														(raw_reason == SNGSS7_REASON_LSSU_SIN) ?"SIN received in in-service state":
														(raw_reason == SNGSS7_REASON_CTS_LOST) ?"Disconnect from L1":
														(raw_reason == SNGSS7_REASON_DAT_IN_OOS) ?"Request to transmit data in OOS":
														(raw_reason == SNGSS7_REASON_DAT_IN_WAITFLUSHCONT) ?"Request to transmit when MTP2 is waiting for flush/continue directive":
														(raw_reason == SNGSS7_REASON_RETRV_IN_INS) ?"Received message retrieval request while in-service state":
														(raw_reason == SNGSS7_REASON_CON_IN_INS) ?"Received Connect request in in-service state":
														(raw_reason == SNGSS7_REASON_UPPER_SAP_DIS) ?"Received management request to disable SAP" : "Unknown");
											}
											break;

										case FTDM_SIG_STATE_SUSPENDED:
											switch (sigmsg->ev_data.sigstatus.raw_reason) {
												case SNGSS7_REASON_START_FLC:                   /* Start of flow control */
													ftdm_log(FTDM_LOG_INFO, "s%dc%d: Flow control start\n", span_id, chan_id);
													chan->flags |= LINK_STATE_CONGESTED;
													break;
												case SNGSS7_REASON_REM_PROC_DN:                 /* Remote MTP2 processor is down */
													ftdm_log(FTDM_LOG_INFO, "s%dc%d: Remote processor down\n", span_id, chan_id);
													break;
												default:
													ftdm_log(FTDM_LOG_INFO, "s%dc%d: Invalid reason for link change state\n", span_id, chan_id);
											}
											break;
										default:
											break;
									}
								}
								break;

							default:
								break;
						}
						/* sample app owns the event buffer. It must be deallocated here
						   otherwise there is possibility of memory leak */
						ftdm_safe_free(sigmsg->raw.data);
						ftdm_safe_free(sigmsg);
					}
				}
				break;

			case FTDM_TIMEOUT:
				/* twiddle */
				break;

			case FTDM_FAIL:
				ftdm_log(FTDM_LOG_ERROR, "ftdm_interrupt_wait returned error!\n");
				break;

			default:
				ftdm_log(FTDM_LOG_ERROR, "ftdm_interrupt_wait returned with unknown code\n");
				break;
		} /* switch (ret_status) */
	} /* while */

exit:
	ftdm_log(FTDM_LOG_CRIT, "app_freetdm_event_queue_handler monitor thread stopping due to error.\n");
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : main()
 *
 * Desc  : Configures all ports and spans as required by user and provide required
 * 	   platform for executing mtp2 api's
 *
 * Ret   : err | FTDM_SUCCESS
**********************************************************************************/
int main(void)
{
	/* local variables initialization */
	int max_wait = 20;
	int err;
	int detected = 0;
	int configured = 0;
	sangoma_hw_config_reg_t reg;

	memset(&reg, 0, sizeof(reg));
	memset(&sng_isup_conf, 0, sizeof(sng_isup_t));

	/* Initialize global FreeTDM init structure */
	memset(&sample_app_ftdm_init, 0, sizeof(sample_app_ftdm_init));

#ifndef __WINDOWS__
	/* Signal handler in order to handle different signal */
	signal(SIGINT, on_sigint);
	signal(SIGTERM, on_sigint);
	signal(SIGUSR1, on_sigusr);
	signal(SIGUSR2, on_sigusr);
#endif

	/* Set basic debuggin parameters */
#ifdef SDK_DEBUGGING
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_INFO;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_ON_ASSERT;
#else
	sample_app_ftdm_init.logging_level = FTDM_LOG_LEVEL_INFO;
	sample_app_ftdm_init.crash_policy_level = FTDM_CRASH_NEVER;
#endif

	/* ftdm_global_set_default_logger(FTDM_LOG_LEVEL_DEBUG); */
	ftdm_global_set_default_logger(FTDM_LOG_LEVEL_INFO);
	ftdm_global_set_crash_policy(FTDM_CRASH_ON_ASSERT);

	ss7_configs = create_hashtable(16, ftdm_hash_hashfromstring, ftdm_hash_equalkeys);

	/* Initilize FreeTDM Framework 
	   Global Varialbes and FreeTDM Logger */
	err = sangoma_freetdm_init(&sample_app_ftdm_init);
	if (err) {
		return err;
	}

	/* Initialize Sangoma Config SDK */
	reg.user_obj = NULL;
	reg.hwprobe_callback = sample_app_hwprobe_callback;
	reg.hwprobe_callback_post = sample_app_span_start_callback;
	err = sangoma_config_reg(&reg);
	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error: Sangoma Config SDK Initialization Error\n");
		exit_sample_app(max_wait);
	}

	/* Create a global FreeTDM event handling queue.
	 * FreeTDM event callback function is not allowed to exectue FreeTDM API commands.
	 * FreeTDM event callback fucntion will queue events, and the app event handler thread
	 * will consume and handle FreeTDM events.
	 * The app event handler is allowed to run FreeTDM api commands directly */
	if (ftdm_thread_create_detached(app_freetdm_event_queue_handler, NULL) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create message queue handler thread. \n");
		exit_sample_app(max_wait);
	}

	/* Detect Sangoma Ports 
	   For each detected port run the sample_app_hwprobe_callback
	 */
	err = sangoma_config_hwprobe(&detected, &configured, SNG_SS7_MTP2_API_MODE_LSL);

	ftdm_log(FTDM_LOG_INFO, "Sangoma SDK Detected %d Configured %d\n", detected, configured);

	if (err) {
		ftdm_log(FTDM_LOG_ERROR, "Error Failed to detect sangoma hardware \n");

		/* We cannot exit here, as the thread is launched */
		exit_sample_app(max_wait);
	}
	/* At this point all sangoma hw ports have been configured and started.
	   And FreeTDM Spans have been created and started on each hw port */
	ftdm_sleep(1000);

	/* User drive main menu. 
	 * Test features for the Sangoma TDM SDK 
	 */
	ftdm_log(FTDM_LOG_INFO, "Sangoma MTP2 API MODULE IN ACTION\n\n\n");

	test_mtp2_links();
	while (mtp2_api_main_menu()) {
		continue;
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : exit_sample_app()
 *
 * Desc  : Stops all FTDM as well as application thread and deallocate the memory
 * 	   stop all the configured ports and exit from application.
 *
 * Ret   : NULL
**********************************************************************************/
int exit_sample_app(int max_wait)
{

	/* Stop all threads. 
	 * TODO: use join instead */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		ftdm_log(FTDM_LOG_INFO, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	/* Stop FreeTDM Freamwork and all Spans */
	ftdm_global_destroy();

	/* Stop all Sangoma hw ports */
	sangoma_stop_all_wanpipe();

	ftdm_sleep(1000);

	return FTDM_SUCCESS;
}

/* PUSHKAR CHANGES STARTS For MTP2 API */
/**********************************************************************************
 * Fun   : add_config_list_nodes()
 *
 * Desc  : Add the nodes that needs to be configured
 *
 * Ret   : FTDM_SUCCESS | -1
**********************************************************************************/
static int add_config_list_nodes(ftdm_conf_node_t *rootnode,
		const char *list_name,
		const char *list_element_name,
		const char *sub_list_name,
		const char *sub_list_element_name)
{
	ftdm_conf_node_t *n_list;
	ftdm_conf_node_t *n_element;
	char *element_name = NULL;
	char *var = NULL;
	char *val = NULL;
	int idx = 0;
	int sub_idx = 0;

	if (!list_name) {
		ftdm_log(FTDM_LOG_ERROR, "No list %s found\n", list_name);
		return -1;
	}

	if ((FTDM_SUCCESS != ftdm_conf_node_create(list_name, &n_list, rootnode))) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create %s node\n", list_name);
		return -1;
	}

	for (idx = 0; idx < 10; idx++) {
		if (strcmp(list_element_name, "mtp1_link") == 0) {
			if (strlen(sng_isup_conf.mtp1_link[idx].name) == 0) {
				if (idx == 0) {
					ftdm_log(FTDM_LOG_ERROR, "Failed to created %s node for %s\n", list_element_name, list_name);
					return -1;
				}
				break;
			}

			element_name = sng_isup_conf.mtp1_link[idx].name;
		} else if (strcmp(list_element_name, "mtp2_link") == 0) {
			if (strlen(sng_isup_conf.mtp2_link[idx].name) == 0) {
				if (idx == 0) {
					ftdm_log(FTDM_LOG_ERROR, "Failed to created %s node for %s\n", list_element_name, element_name);
					return -1;
				}
				break;
			}

			element_name = sng_isup_conf.mtp2_link[idx].name;
		}


		if ((FTDM_SUCCESS != ftdm_conf_node_create(list_element_name, &n_element, n_list))) {
			ftdm_log(FTDM_LOG_ERROR, "Failed to create %s node for %s\n", list_element_name, element_name);
			return -1;
		}
		ftdm_conf_node_add_param(n_element, "name", element_name);

		for (sub_idx = 0; sub_idx < 10; sub_idx++) {
			if (strcmp(list_element_name, "mtp1_link") == 0) {
				if (strlen(sng_isup_conf.mtp1_link[idx].p_name[sub_idx]) == 0) {
					break;
				}
				var = sng_isup_conf.mtp1_link[idx].p_name[sub_idx];
				val = sng_isup_conf.mtp1_link[idx].p_value[sub_idx];
			} else if (strcmp(list_element_name, "mtp2_link") == 0) {
				if (strlen(sng_isup_conf.mtp2_link[idx].p_name[sub_idx]) == 0) {
					break;
				}
				var = sng_isup_conf.mtp2_link[idx].p_name[sub_idx];
				val = sng_isup_conf.mtp2_link[idx].p_value[sub_idx];
			}

			ftdm_conf_node_add_param(n_element, var, val);
		}

		if (sub_list_name && sub_list_element_name) {
			if (add_config_list_nodes(n_element, sub_list_name, sub_list_element_name, NULL, NULL)) {
				return -1;
			}
		}
	}

	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : get_ss7_config_node()
 *
 * Desc  : Get the node that needs to be configured as per customers configuration
 *
 * Ret   : NULL | rootnode
**********************************************************************************/
static ftdm_conf_node_t *get_ss7_config_node(char *confname)
{
	ftdm_conf_node_t *rootnode, *list;
	char *var = NULL;
	char *val = NULL;
	int idx = 0;

	/* try to find the conf in the hash first */
	rootnode = hashtable_search(ss7_configs, (void *)confname);
	if (rootnode) {
		ftdm_log(FTDM_LOG_DEBUG, "SS7 config %s was found in the hash already\n", confname);
		return rootnode;
	}
	ftdm_log(FTDM_LOG_DEBUG, "Not found %s config in hash, searching in xml ...\n", confname);

	char *name = sng_isup_conf.name;

	if (strcasecmp(name, confname)) {
		ftdm_log(FTDM_LOG_ERROR, "sng_isup name[%s] does not match with configprofile[%s]\n", name, confname);
		return NULL;
	}

	/* create the root node */
	if ((FTDM_SUCCESS != ftdm_conf_node_create("sng_isup", &rootnode, NULL))) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create root node for sng_isup config %s\n", confname);
		return NULL;
	}

	/* add sng_gen */
	if (strlen(sng_isup_conf.sng_gen.p_name[0]) == 0) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to process sng_gen for sng_isup config %s\n", confname);
		ftdm_conf_node_destroy(rootnode);
		return NULL;
	}

	if ((FTDM_SUCCESS != ftdm_conf_node_create("sng_gen", &list, rootnode))) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create %s node for %s\n", "sng_gen", confname);
		ftdm_conf_node_destroy(rootnode);
		return NULL;
	}

	for (idx = 0; idx < 10; idx++) {
		if (strlen(sng_isup_conf.sng_gen.p_name[idx]) == 0) {
			if (idx == 0) {
				ftdm_log(FTDM_LOG_ERROR, "Failed to create %s node for %s\n", "sng_gen", confname);
				ftdm_conf_node_destroy(rootnode);
				return NULL;
			}
			break;
		}
		var = sng_isup_conf.sng_gen.p_name[idx];
		val = sng_isup_conf.sng_gen.p_value[idx];
		ftdm_conf_node_add_param(list, var, val);
	}

	/* add mtp1 links */
	if (add_config_list_nodes(rootnode, "mtp1_links", "mtp1_link", NULL, NULL)) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to process mtp1_links for sng_isup config %s\n", confname);
		ftdm_conf_node_destroy(rootnode);
		return NULL;
	}

	/* add mtp2 links */
	if (add_config_list_nodes(rootnode, "mtp2_links", "mtp2_link", NULL, NULL)) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to process mtp2_links for sng_isup config %s\n", confname);
		ftdm_conf_node_destroy(rootnode);
		return NULL;
	}

	hashtable_insert(ss7_configs, confname, rootnode, HASHTABLE_FLAG_NONE);

	ftdm_log(FTDM_LOG_DEBUG, "Added SS7 node configuration %s\n", confname);
	return rootnode;
}

/**********************************************************************************
 * Fun   : config_spans()
 *
 * Desc  : Detect the port number and configure span as per user requirement.
 *
 * Ret   : FTDM_SUCCESS | FTDM_FAIL | -1
**********************************************************************************/
int config_spans(void *sangoma_port_obj, int wanpipe_num, span_config_t * span_conf)
{
	char *configname = NULL;
	int card_type = sangoma_config_port_hwinfo_get_model_id(sangoma_port_obj);
	int port_num = sangoma_config_port_hwinfo_get_port_number(sangoma_port_obj);

	memset(span_conf, 0, sizeof(span_config_t));

	switch (card_type) {
		case A101_ADPTR_2TE1:
		case A104_ADPTR_4TE1:
		case A108_ADPTR_8TE1:
		case A116_ADPTR_16TE1:
			{
				switch (port_num) {
					case 1:
						/* sangoma ss7 span configuration */
						strcpy(span_conf->config.mtp2_api.span_name, "wp13");
						strcpy(span_conf->config.mtp2_api.cfgprofile, "ss7config");
						strcpy(span_conf->config.mtp2_api.p_name[0], "ccSpanId");
						strcpy(span_conf->config.mtp2_api.p_value[0], "1");

						/* sangoma ss7 nodes i.e. mtp1/mtp2/isup etc. configuration */
						strcpy(sng_isup_conf.name, "ss7config");

						/* sangoma ss7 general configuration */
						strcpy(sng_isup_conf.sng_gen.p_name[0], "procId");
						strcpy(sng_isup_conf.sng_gen.p_value[0], "1");
						strcpy(sng_isup_conf.sng_gen.p_name[1], "operating_mode");
						strcpy(sng_isup_conf.sng_gen.p_value[1], "MTP2_API");
						strcpy(sng_isup_conf.sng_gen.p_name[2], "license");
						strcpy(sng_isup_conf.sng_gen.p_value[2], "/usr/local/nsg/conf/license.txt");

						/* sangoma ss7 mtp1 link configuration */
						/* 1st mtp1 link config*/
						strcpy(sng_isup_conf.mtp1_link[0].name, "mtp1-1");
						strcpy(sng_isup_conf.mtp1_link[0].p_name[0], "id");
						strcpy(sng_isup_conf.mtp1_link[0].p_value[0], "1");
						strcpy(sng_isup_conf.mtp1_link[0].p_name[1], "span_name");
						strcpy(sng_isup_conf.mtp1_link[0].p_value[1], "wp13");
						strcpy(sng_isup_conf.mtp1_link[0].p_name[2], "chan_no");
						strcpy(sng_isup_conf.mtp1_link[0].p_value[2], "1");

						/* 2nd mtp1 link config*/
						strcpy(sng_isup_conf.mtp1_link[1].name, "mtp1-2");
						strcpy(sng_isup_conf.mtp1_link[1].p_name[0], "id");
						strcpy(sng_isup_conf.mtp1_link[1].p_value[0], "2");
						strcpy(sng_isup_conf.mtp1_link[1].p_name[1], "span_name");
						strcpy(sng_isup_conf.mtp1_link[1].p_value[1], "wp14");
						strcpy(sng_isup_conf.mtp1_link[1].p_name[2], "chan_no");
						strcpy(sng_isup_conf.mtp1_link[1].p_value[2], "1");

						/* sangoma ss7 mtp2 link configuration */
						/* 1st mtp2 link config*/
						strcpy(sng_isup_conf.mtp2_link[0].name, "mtp2-1");
						strcpy(sng_isup_conf.mtp2_link[0].p_name[0], "id");
						strcpy(sng_isup_conf.mtp2_link[0].p_value[0], "1");
						strcpy(sng_isup_conf.mtp2_link[0].p_name[1], "mtp1Id");
						strcpy(sng_isup_conf.mtp2_link[0].p_value[1], "1");
						strcpy(sng_isup_conf.mtp2_link[0].p_name[2], "errorType");
						strcpy(sng_isup_conf.mtp2_link[0].p_value[2], "basic");
						strcpy(sng_isup_conf.mtp2_link[0].p_name[3], "lssuLength");
						strcpy(sng_isup_conf.mtp2_link[0].p_value[3], "1");
						strcpy(sng_isup_conf.mtp2_link[0].p_name[4], "linkType");
						strcpy(sng_isup_conf.mtp2_link[0].p_value[4], "itu92");
						strcpy(sng_isup_conf.mtp2_link[0].p_name[5], "sapType");
						strcpy(sng_isup_conf.mtp2_link[0].p_value[5], "lsl");

						/* 2nd mtp2 link config*/
						strcpy(sng_isup_conf.mtp2_link[1].name, "mtp2-2");
						strcpy(sng_isup_conf.mtp2_link[1].p_name[0], "id");
						strcpy(sng_isup_conf.mtp2_link[1].p_value[0], "2");
						strcpy(sng_isup_conf.mtp2_link[1].p_name[1], "mtp1Id");
						strcpy(sng_isup_conf.mtp2_link[1].p_value[1], "2");
						strcpy(sng_isup_conf.mtp2_link[1].p_name[2], "errorType");
						strcpy(sng_isup_conf.mtp2_link[1].p_value[2], "basic");
						strcpy(sng_isup_conf.mtp2_link[1].p_name[3], "lssuLength");
						strcpy(sng_isup_conf.mtp2_link[1].p_value[3], "1");
						strcpy(sng_isup_conf.mtp2_link[1].p_name[4], "linkType");
						strcpy(sng_isup_conf.mtp2_link[1].p_value[4], "itu92");
						strcpy(sng_isup_conf.mtp2_link[1].p_name[5], "sapType");
						strcpy(sng_isup_conf.mtp2_link[1].p_value[5], "lsl");

						break;

					case 2:
						/* sangoma ss7 span configuration */
						strcpy(span_conf->config.mtp2_api.span_name, "wp14");
						strcpy(span_conf->config.mtp2_api.cfgprofile, "ss7config");
						strcpy(span_conf->config.mtp2_api.p_name[0], "ccSpanId");
						strcpy(span_conf->config.mtp2_api.p_value[0], "2");
						break;

					case 3:
					case 4:
					case 5:
					case 6:
					case 7:
					case 8:
						break;

					default:
						ftdm_log(FTDM_LOG_ERROR, "Invalid port id = %d\n", port_num);
						return -1;
						break;
				}
			}
			break;

		default:
			ftdm_log(FTDM_LOG_ERROR, "%s card (%d) does not support MTP2 API\n", SNG_MTP2_API_CARD_TYPE(card_type),
					card_type);
			return FTDM_FAIL;
	}

	span_conf->config.mtp2_api.ss7confnode = NULL;
	configname = span_conf->config.mtp2_api.cfgprofile;
	span_conf->config.mtp2_api.ss7confnode = get_ss7_config_node(configname);

	return FTDM_SUCCESS;
}
