/*
 * Copyright (c) 2014,
 * Sangoma Technologies
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * * Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *
 * * Neither the name of the original author; nor the names of any contributors
 *   may be used to endorse or promote products derived from this software
 *   without specific prior written permission.
 *
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER
 * OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Contributors: Pushkar Singh <psingh@sangoma.com>
 * 		 David Yat Sin <dyatsin@sangoma.com>
 *
 */

/********************************************************************************/
/*                                                                              */
/*                           libraries inclusion                                */
/*                                                                              */
/********************************************************************************/
#include "sample_mtp2_api.h"

/********************************************************************************/
/*                                                                              */
/*                             MACROS declaration                               */
/*                                                                              */
/********************************************************************************/
#define STACK_SIZE 1024*1024
#define MIN_MSU_SIZE 8

/********************************************************************************/
/*                                                                              */
/*                           function declaration                               */
/*                                                                              */
/********************************************************************************/
extern span_data_t SPANS[FTDM_MAX_SPANS_INTERFACE];
extern int g_running;
extern int g_num_threads;

static void *mtp2_chan_run (ftdm_thread_t *me, void* chan_ptr);
void verify_receive_data(chan_data_t *chan, void *vdata, uint32_t datalen);
int exit_mtp2_api(int max_wait);

/********************************************************************************/
/*                                                                              */
/*                              global parameters                               */
/*                                                                              */
/********************************************************************************/

/**********************************************************************************
 * Fun   : printf_frames(void)
 *
 * Desc  : Print information w.r.t number of frames recieved & transmitted
 *
 * Ret   : void
***********************************************************************************/
void print_frames(void)
{
	int i,j;
	unsigned print_buf_len = 0;
	char print_buf [10000];
	memset(print_buf, 0, 10000);
	
	print_buf_len += sprintf(&print_buf[print_buf_len], "\n======================================================================================================================================\n");

	print_buf_len += sprintf(&print_buf[print_buf_len], "Format: <Timeslot>: <Number of Tx Frames>/<Number of Rx Frames>");
	for (i = 0; i < FTDM_MAX_SPANS_INTERFACE; i++) {
		if (SPANS[i].inuse) {
			print_buf_len += sprintf(&print_buf[print_buf_len], "\nSpan :%d\n\t",i);
			for (j = 0; j < NUM_E1_CHANS; j++) {
				if (SPANS[i].channels[j].inuse) {
					print_buf_len += sprintf(&print_buf[print_buf_len], "%02d:%06ld/%06ld    ", j, SPANS[i].channels[j].tx_frames, SPANS[i].channels[j].rx_frames);
					if (j && !((j+1)%6)) print_buf_len += sprintf(&print_buf[print_buf_len], "\n\t");
				}
			}
		}
	}	
	print_buf_len += sprintf(&print_buf[print_buf_len], "\n======================================================================================================================================\n");
	ftdm_log(FTDM_LOG_INFO, "%s\n", print_buf);
	return;
}

/**********************************************************************************
 * Fun   : printf_status(void)
 *
 * Desc  : Print the status w.r.t the spans configured. i.e. whether the span is in
 * 	   Aligned/Suspended/Disconnected state
 *
 * Ret   : void
***********************************************************************************/
void print_status(void)
{
	int i,j;
	unsigned print_buf_len = 0;
	char print_buf [1000];
	memset(print_buf, 0, 1000);
	
	print_buf_len += sprintf(&print_buf[print_buf_len], "\n===========================================================================================================================================================\n");
	for (i = 0; i < FTDM_MAX_SPANS_INTERFACE; i++) {
		if (SPANS[i].inuse) {
			print_buf_len += sprintf(&print_buf[print_buf_len], "\nSpan :%d\n\t",i);
			for (j = 0; j < NUM_E1_CHANS; j++) {
				if (SPANS[i].channels[j].inuse) {
					ftdm_signaling_status_t sigstatus;
					if (ftdm_channel_get_sig_status(SPANS[i].channels[j].ftdmchan, &sigstatus) != FTDM_SUCCESS) {
						ftdm_log(FTDM_LOG_ERROR, "s%dc%d: Failed to get signalling status on channel\n", SPANS[i].channels[j].span_id, SPANS[i].channels[j].chan_id);
					} else {
						print_buf_len += sprintf(&print_buf[print_buf_len], "%s \t", (sigstatus == FTDM_SIG_STATE_UP)?"Aligned":
													     (sigstatus == FTDM_SIG_STATE_SUSPENDED)?"Suspended":
													     (sigstatus == FTDM_SIG_STATE_DOWN)?"Disconnected":"Unknown");
					}
					
					if (j && !((j+1)%6)) print_buf_len += sprintf(&print_buf[print_buf_len], "\n\t");
				}
			}
		}
	}
	print_buf_len += sprintf(&print_buf[print_buf_len], "\n===========================================================================================================================================================\n");
	ftdm_log(FTDM_LOG_INFO, "%s\n", print_buf);
	return;
}

/**********************************************************************************
 * Fun   : launch_test_thread(chan_data_t *)
 *
 * Desc  : Created/Launch a separate thread in order to recieve and tranmit data
 *
 * Ret   : int
***********************************************************************************/
static int launch_test_thread(chan_data_t *chan)
{
	if (ftdm_thread_create_detached(mtp2_chan_run,(void *)chan) != FTDM_SUCCESS) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to create thread!!\n");
		return FTDM_FAIL;
	}
	return FTDM_SUCCESS;
}

/**********************************************************************************
 * Fun   : test_mtp2_links(void)
 *
 * Desc  : Launches the test thread in order to send data to other MTP2 span and
 * 	   test MTP2 api functionality
 *
 * Ret   : void
***********************************************************************************/
void test_mtp2_links(void)
{
	int i, j=0;
	ftdm_channel_t *ftdmchan = NULL;
	ftdm_iterator_t *chaniter = NULL;
	ftdm_iterator_t *curr = NULL;
	
	for (i = 0; i < FTDM_MAX_SPANS_INTERFACE; i++) {
		if (SPANS[i].inuse) {
			j = 0;
			chaniter = ftdm_span_get_chan_iterator(SPANS[i].ftdmspan, chaniter);
			for (curr = chaniter; curr; curr = ftdm_iterator_next(curr)) {
				ftdmchan = ftdm_iterator_current(curr);
				SPANS[i].channels[j].span_id = ftdm_channel_get_ph_span_id(ftdmchan);
				SPANS[i].channels[j].chan_id = ftdm_channel_get_ph_id(ftdmchan);
				SPANS[i].channels[j].ftdmchan = ftdmchan;
				SPANS[i].channels[j].inuse = 1;

				ftdm_log(FTDM_LOG_INFO, "Launching test thread for span:%d channel:%d\n", SPANS[i].channels[j].span_id, SPANS[i].channels[j].chan_id);

				if (launch_test_thread(&SPANS[i].channels[j])) {
					g_running = 0;
					ftdm_iterator_free(chaniter);
					return;
				}

				j++;
				g_num_threads++;
			}
		}
	}
	ftdm_iterator_free(chaniter);
	ftdm_log(FTDM_LOG_INFO, "Number of tests launched:%d\n", g_num_threads);
}

/**********************************************************************************
 * Fun   : request_mtp2_connect(ftdm_channel_t *)
 *
 * Desc  : Request MTP2 to start alignment attempt
 *
 * Ret   : void
***********************************************************************************/
static void request_mtp2_connect(ftdm_channel_t *ftdmchan)
{
	ftdm_usrmsg_t usrmsg;
	
	memset(&usrmsg, 0, sizeof(usrmsg));

	/* usrmsg.raw_id = SNGSS7_SIGSTATUS_UP_NORM; */
	usrmsg.raw_id = SNGSS7_SIGSTATUS_EMERG;

	ftdm_channel_set_sig_status_ex(ftdmchan, FTDM_SIG_STATE_UP,  &usrmsg);
	return;
}

/**********************************************************************************
 * Fun   : request_mtp2_flc_status(ftdm_channel_t *)
 *
 * Desc  : Check MTP2 flow control status
 *
 * Ret   : void
***********************************************************************************/
static void request_mtp2_flc_status(ftdm_channel_t *ftdmchan)
{
	ftdm_usrmsg_t usrmsg;
	
	memset(&usrmsg, 0, sizeof(usrmsg));

	/* usrmsg.raw_id = SNGSS7_SIGSTATUS_UP_NORM; */
	usrmsg.raw_id = SNGSS7_SIGSTATUS_FLC_STAT;

	ftdm_channel_set_sig_status_ex(ftdmchan, FTDM_SIG_STATE_UP,  &usrmsg);
	return;
}

/**********************************************************************************
 * Fun   : mtp2_chan_run (ftdm_thread_t *, void *)
 *
 * Desc  : Opens the signalling channel of respective span and start sending data
 * 	   to other MTP2 channel connected
 *
 * Ret   : void
***********************************************************************************/
static void *mtp2_chan_run (ftdm_thread_t *me, void* chan_ptr)
{
	unsigned int span_id;
	unsigned int chan_id;
	ftdm_channel_t *ftdmchan;
	uint8_t *data;
	ftdm_size_t datalen;
	uint32_t sleeptime;
	uint8_t transmit_data = 0x00;
	chan_data_t *chan = (chan_data_t*)chan_ptr;
	ftdm_status_t status = FTDM_SUCCESS;

	data = malloc(MAX_MSU_SIZE);
	if (!data) {
		ftdm_log(FTDM_LOG_ERROR, "Failed to allocate memory\n");
		exit_mtp2_api(20);
	}

	span_id = chan->span_id;
	chan_id = chan->chan_id;

	ftdm_log(FTDM_LOG_INFO, "Opening channel (span:%d chan:%d)\n", span_id, chan_id);
	status = ftdm_channel_open(span_id, chan_id, &ftdmchan);

	if (!ftdmchan || (status != FTDM_SUCCESS && status != FTDM_EBUSY)) {
		ftdm_log(FTDM_LOG_INFO,"-ERR Failed to open channel %d in span %d\n", chan_id, span_id);
		if(FTDM_EBUSY == status)
		{
			ftdm_log(FTDM_LOG_INFO,"Channel already opened \n");
		}
		exit_mtp2_api(20);
	}


	ftdm_channel_set_private(ftdmchan, chan);

	sleep(1);

	/* Transmit dummy MSU's */
	while(g_running) {
		if (!(chan->flags & LINK_STATE_CONNECTED)) {
			ftdm_log(FTDM_LOG_INFO, "s%dc%d: Link is down\n", span_id, chan_id);

			request_mtp2_connect(ftdmchan);			
			sleep(1);
			continue;
		}
		if (chan->flags & LINK_STATE_CONGESTED) {
			ftdm_log(FTDM_LOG_INFO, "s%dc%d: Link in congested state\n", span_id, chan_id);
			request_mtp2_flc_status(ftdmchan);			
			
			/* Cannot transmit when link is in congested state, wait for congested condition to clear */
			usleep(10000);
			continue;
		}
		datalen = (uint32_t)((float)rand()/(float)RAND_MAX*(float)(MAX_MSU_SIZE));

		if (datalen > MIN_MSU_SIZE) {
			datalen -= 1; /* So that datalen is never == MAX_MSU_SIZE; */

			/* Fill in the dummy buffer with some information so we can verify data when the MSU is received */
			data[0] = span_id;
			data[1] = chan_id;
			data[2] = (datalen & 0xFF);
			data[3] = ((datalen >> 8) & 0xFF);

			memset(&data[4], transmit_data++, (datalen-4));

			if (ftdm_channel_write(ftdmchan, data, MAX_MSU_SIZE, &datalen) != FTDM_SUCCESS) {
				ftdm_log(FTDM_LOG_ERROR, "Failed to write to channel\n");
				break;
			}
			chan->tx_frames++;

			/* Sleep, otherwise we will enter congestion state each time because we would transmit
			 * as fast as we can in the sample application. Sleep is not required in a real application */

			/* We transmit 8 bytes in 1 ms, so it takes datalen/8 ms to transmit a MSU + ~10 bytes for FISU*/
			sleeptime = (uint32_t)(((float)datalen/(float)8)+10);			
			usleep(sleeptime*1000);
		}
	}

	ftdm_channel_close(&ftdmchan);
	free(data);
	g_num_threads--;	
	exit_mtp2_api(20);
}

/**********************************************************************************
 * Fun   : verify_receive_data(chan_data_t *, void *, uint32_t)
 *
 * Desc  : Verified the data recieved whether it is same that is being transmitted
 *	   or not
 *
 * Ret   : void
***********************************************************************************/
void verify_receive_data(chan_data_t *chan, void *vdata, uint32_t datalen)
{

	uint32_t rx_span_id, rx_chan_id, rx_datalen;
	uint8_t *data = (uint8_t *)vdata;
	uint8_t rx_seq;

	if (datalen < MIN_MSU_SIZE) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid MSU length received %d\n", datalen);
		g_running = 0;
		return;
	}

	rx_span_id = data[0];
	rx_chan_id = data[1];
	rx_datalen = data[2];
	rx_datalen |= ((data[3] & 0xFF) << 8);
	rx_seq		= data[4];

#if 0
	if (chan->span_id != rx_span_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid data: span id RX:%d expected: %d\n", rx_span_id, chan->span_id);
		g_running = 0;
		return;
	}
#endif

	if (chan->chan_id != rx_chan_id) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid data: chan id RX:%d expected: %d\n", rx_chan_id, chan->chan_id);
		g_running = 0;
		return;
	}
	
	if (datalen != rx_datalen) {
		ftdm_log(FTDM_LOG_ERROR, "Invalid data: Data Length RX:%d expected: %d\n", rx_datalen, datalen);
		g_running = 0;
		return;
	}

	if (chan->rx_seq != rx_seq) {
		ftdm_log(FTDM_LOG_ERROR, "Dropped data: RX Seq:0x%02x expected Seq:0x%02x \n", rx_seq, chan->rx_seq);
		g_running = 0;
		return;
	}
	
	chan->rx_seq = rx_seq+1;
	chan->rx_frames++;
}

/**********************************************************************************
 * Fun   : exit_sample_app()
 *
 * Desc  : Stops all FTDM as well as application thread and deallocate the memory
 *         stop all the configured ports and exit from application.
 *
 * Ret   : NULL
 **********************************************************************************/
int exit_mtp2_api(int max_wait)
{
	/* Stop all threads */
	g_running = 0;
	ftdm_sleep(1000);

	while (g_num_threads && (max_wait-- > 0)) {
		ftdm_log(FTDM_LOG_INFO, "Waiting for all threads to finish (%d)\n", g_num_threads);
		ftdm_sleep(100);
	}

	return FTDM_FAIL;
}
